#include <avr/io.h>
#include <util/delay.h>

#define POWER PC0 // Active LOW
#define FAIL PC1
#define SUCCESS PC2
#define XTAL PC3
#define HV_DISABLE PC4
#define START PC5

#define BS2 PD0
#define RDY PD1
#define OE PD2
#define WR PD3
#define BS1 PD4
#define XA0 PD5
#define XA1 PD6
#define PAGEL PD7

#define DATA_PORT PORTB
#define DATA_DIR DDRB

#define LFUSE 0xC1
#define HFUSE 0x99

const uint8_t signature[3] = {0x1E, 0x93, 0x07};

void load_cmd(uint8_t cmd){
    // XA1, XA0 = 10; BS1 = 0
    PORTD |= (1 << XA1);
    PORTD &= ~((1 << XA0) | (1 << BS1));

    PORTB = cmd;
    DDRB = 0xff;

    // Clcok XTAL
    PORTC |= (1 << XTAL);
    _delay_us(1);
    PORTC &= ~(1 << XTAL);
    _delay_us(1);

    DDRB = 0;
}

void load_addr(uint8_t addr){
    // XA1, XA0 = 00; BS1 = 0
    PORTD &= ~((1 << XA1) | (1 << XA0) | (1 << BS1));

    PORTB = addr;
    DDRB = 0xff;

    // Clcok XTAL
    PORTC |= (1 << XTAL);
    _delay_us(1);
    PORTC &= ~(1 << XTAL);
    _delay_us(1);

    DDRB = 0;
}

void load_data(uint8_t data){
    // XA1, XA0 = 01; BS1 = 0
    PORTD |= (1 << XA0);
    PORTD &= ~((1 << XA1) | (1 << BS1));

    PORTB = data;
    DDRB = 0xff;

    // Clcok XTAL
    PORTC |= (1 << XTAL);
    _delay_us(1);
    PORTC &= ~(1 << XTAL);
    _delay_us(1);

    DDRB = 0;
}

uint8_t verify_signature(){
    load_cmd(0b00001000); // Read SIG BYTES cmd
    
    for(uint8_t i = 0; i < 0x03; i++){
        load_addr(i);
        _delay_us(1);

        DDRB = 0; // We are reading inputs only
        PORTD &= ~(1 << OE); // Enable the data
        _delay_us(100);
        
        if(PINB != signature[i]){
            return 0;
        }

        PORTD |= (1 << OE); // OUTPUT again
        _delay_us(1);
    }

    return 1;
}

void chip_erase() {
    load_cmd(0b10000000); // Chip erase cmd

    // Pulse WR
    PORTD &= ~(1 << WR);
    _delay_us(1);
    PORTD |= (1 << WR);
    _delay_us(1);

    while(!(PIND & (1 << RDY))); // Wait until device is done

    _delay_us(1);
}

void program_fuses(uint8_t low, uint8_t high){
    load_cmd(0b01000000); // Write fuse bits cmd

    // Start of LOW byte
    load_data(low);
    PORTD &= ~((1 << BS1) | (1 << BS2));

    // Pulse WR
    PORTD &= ~(1 << WR);
    _delay_us(1);
    PORTD |= (1 << WR);
    _delay_us(1);

    while(!(PIND & (1 << RDY))); // Wait until device is done

    // Start of HIGH byte
    load_data(high);
    PORTD |= (1 << BS1);

    // Pulse WR
    PORTD &= ~(1 << WR);
    _delay_us(1);
    PORTD |= (1 << WR);
    _delay_us(1);

    while(!(PIND & (1 << RDY))); // Wait until device is done

    PORTD &= ~(1 << BS1);
}

uint8_t verify_fuses() {
    load_cmd(0b00000100); // Read fuses and lock bits cmd
    DDRB = 0; // Reading only

    // LFUSE
    PORTD &= ~((1 << BS1) | (1 << BS2));
    PORTD &= ~(1 << OE);
    _delay_us(1);
    if(PINB != LFUSE){
        return 0;
    }

    // LOCK
    PORTD |= (1 << BS1);
    _delay_us(1);
    if(PINB != 0xFF){
        return 0;
    }

    // HFUSE
    PORTD |= (1 << BS2);
    _delay_us(1);
    if(PINB != HFUSE){
        return 0;
    }

    PORTD |= (1 << OE);
    return 1;
}

void reset_programming(){
    PORTC |= (1 << HV_DISABLE) | (1 << POWER); // 12 Volts & POWER off
    PORTC &= ~(1 << XTAL); // XTAL off

    PORTD = (1 << OE) | (1 << WR); // OE and WR are active LOW
    DDRD = 0; // High-Z
    DDRB = 0; // High-Z
}

uint8_t start_programming(){
    // Try method 1 first
    PORTC &= ~(1 << POWER);
    _delay_us(110);
    PORTC |= (1 << HV_DISABLE); // Keep RESET at 0

    // Toggle XTAL at least 6 times
    for(uint8_t i = 0; i < 6; i++){ 
        PORTC |= (1 << XTAL); 
        _delay_us(1);
        PORTC &= ~(1 << XTAL);
    }
    _delay_us(1);

    // Prog_enable pins
    PORTD &= ~((1 << BS1) | (1 << PAGEL) | (1 << XA0) | (1 << XA1));
    DDRD |= (1 << BS1) | (1 << PAGEL) | (1 << XA0) | (1 << XA1);
    _delay_us(1);

    PORTC &= ~(1 << HV_DISABLE); // Enable +12V
    _delay_us(100);

    if(PIND & (1 << RDY)) // Success: continue to programming
        return 1;

    // Failure: METHOD 2
    reset_programming();
    _delay_us(100);

    // Prog_enable pins
    PORTD &= ~((1 << BS1) | (1 << PAGEL) | (1 << XA0) | (1 << XA1));
    DDRD |= (1 << BS1) | (1 << PAGEL) | (1 << XA0) | (1 << XA1);

    // Enable POWER and HV simultaneously
    PORTC &= ~(1 << POWER);
    PORTC &= ~(1 << HV_DISABLE);
    _delay_us(1);

    if(!(PIND & (1 << RDY))) // Not in programming mode
        return 0;

    return 1;
}

int main() {
    // Setup inputs and outputs
    PORTC = (1 << HV_DISABLE) | (1 << START) | (1 << POWER); // 12 Volts Off, Pullup on START, POWER off
    DDRC = (1 << POWER) | (1 << FAIL) | (1 << SUCCESS) | (1 << XTAL) | (1 << HV_DISABLE);

    PORTD = (1 << OE) | (1 << WR); // OE and WR are active LOW

    while(1){
        _delay_ms(500);
        reset_programming();
        PORTC &= ~((1 << FAIL) | (1 << SUCCESS));

        while(PINC & (1 << START)); // Wait until Start button is pressed

        if(!start_programming()){ // If error, stop process
            PORTC |= (1 << FAIL);
            continue;
        } 

        DDRD = ~(1 << RDY); // Full Program control

        if(!verify_signature()){
            PORTC |= (1 << FAIL);
            continue;
        }

        chip_erase(); // Reset Lock Bits

        program_fuses(LFUSE, HFUSE);

        if(!(verify_fuses())){
            PORTC |= (1 << FAIL);
            continue;
        }

        PORTC |= (1 << SUCCESS);
    }

}
